{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Homework 5\n",
    "## Due Thursday, November 7th 2019 at 11:59 PM.\n",
    "\n",
    "[Problem 0](#p0).  Homework Workflow [10 pts]\n",
    "\n",
    "[Problem 1](#p1).  `@property` Decorator [20 points]\n",
    "\n",
    "[Problem 2](#p2).  Linked List [40 pts]\n",
    "\n",
    "[Problem 3](#p3).  Binary Search Tree [30 pts]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**We will be using an autograder for this homework. Please don't change the specified function signature.**"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<a id='p0'></a>\n",
    "## Problem 0: Homework Workflow\n",
    "Once you receive HW4 feedback, you will need to merge your `HW4-dev` branch into `master`. \n",
    "\n",
    "You will earn points for following all stages of the git workflow which involves:\n",
    "* 3pts for merging `HW4-dev` into `master`\n",
    "* 5pts for completing HW5 on `HW5-dev`\n",
    "* 2pts for making a PR on `HW5-dev` to merge into `master`"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<a id='p1'></a>\n",
    "## Problem 1: `@property` Decorator\n",
    "\n",
    "Suppose we have a simple `Animal` class, which is initialized by its species:\n",
    "\n",
    "```python\n",
    "class Animal:\n",
    "    \n",
    "    def __init__(self, name, species):\n",
    "        self.name = name\n",
    "        self.species = species\n",
    "        \n",
    "    def __repr__(self):\n",
    "        return f'{self.name} ({self.species})'\n",
    "```\n",
    "\n",
    "We have clients who build applications on top of our `Animal` class and they want to directly modify the `species` attribute. For example, they may implement functions to transform animals.\n",
    "\n",
    "```python\n",
    "def transform(animal, into):\n",
    "    animal.species = into\n",
    "    return animal\n",
    "```\n",
    "\n",
    "However, one day it occurred to us that our `Animal` class is designed to be used in an [H. C. Andersen universe](https://en.wikipedia.org/wiki/Hans_Christian_Andersen#Fairy_tales_and_poetry). We really don't want people to make an `Animal` instance into a [Cthulhu](https://en.wikipedia.org/wiki/Cthulhu) [Cthulhu is not part of the H. C. Anderson universe]. How should we enforce this restriction? The obvious solution is to create a method to set the species, which checks if the new species is valid inside the method:\n",
    "\n",
    "```python\n",
    "class Animal:\n",
    "    \n",
    "    # a class attribute of the valid species in our universe\n",
    "    valid_species = {\n",
    "        'cat',\n",
    "        'dog',\n",
    "        'duck',\n",
    "        'elf',\n",
    "        'goblin',\n",
    "        'horse',\n",
    "        'human',\n",
    "        'mermaid',\n",
    "        'nightingale',\n",
    "        'pig',\n",
    "        'swan',\n",
    "        'wolf'\n",
    "    }\n",
    "    \n",
    "    def __init__(self, species):\n",
    "        self.species = species\n",
    "            \n",
    "    def __repr__(self):\n",
    "        return f'{self.name} ({self.species})'\n",
    "    \n",
    "    def set_species(self, into):\n",
    "        assert into in Animal.valid_species, Exception(f'invalid species: {into}')\n",
    "        self.species = into\n",
    "```\n",
    "\n",
    "There are two problems with this implementation: \n",
    "1. If our clients don't know they have to change their implementation of `transform` (use `set_species` instead of directly assign the `species` attribute), people can still do things like `a = transform(ugly_duckling, into='Cthulhu')`\n",
    "2. We really don't want to bother our clients to change their code because they may have used `animal.species = something` hundreds of times throughout their code.\n",
    "\n",
    "Fortunately, we have the `property` decorator. Read more about it [here](https://www.programiz.com/python-programming/property). \n",
    "\n",
    "For this problem, you will utilize the `property` decorator to limit what species users can set their `Animal` object to. Please copy the following code into the file `animal.py`.\n",
    "\n",
    "```python\n",
    "class Animal:\n",
    "    \n",
    "    # a class attribute of the valid species in our universe\n",
    "    valid_species = {\n",
    "        'cat',\n",
    "        'dog',\n",
    "        'duck',\n",
    "        'elf',\n",
    "        'goblin',\n",
    "        'horse',\n",
    "        'human',\n",
    "        'mermaid',\n",
    "        'nightingale',\n",
    "        'pig',\n",
    "        'swan',\n",
    "        'wolf'\n",
    "    }\n",
    "    \n",
    "    def __init__(self, name, species):\n",
    "        self.name = name\n",
    "        self._species = species\n",
    "    \n",
    "    def __repr__(self):\n",
    "        return f'{self.name} ({self._species})'\n",
    "```\n",
    "\n",
    "\n",
    "### Part A [10 points]:\n",
    "\n",
    "We have also now made the `species` attribute somewhat private by calling it `_species` because we do not want the user to access it directly. However, this presents a problem because in our first version of the `Animal` class, a client could access the species through the `species` attribute. You can check that the attributes of the `Animal` class are `name` and `_species` using the following steps:\n",
    "```python\n",
    ">>> dog = Animal('Fido', 'dog')\n",
    ">>> vars(dog)\n",
    "{'name': 'Fido',  '_species' : 'dog'}\n",
    "```\n",
    "\n",
    "To prevent the client's previous code from breaking, modify the `Animal` class in `animal.py` by adding a `species(self)` method to the `Animal` class. The `species` method should return the `_species` attribute of this `Animal`. Apply the [`property` decorator](https://www.programiz.com/python-programming/property) to this method.\n",
    "\n",
    "If this is done correctly, then a client should still be able to retrieve the species of an `Animal` object as follows:\n",
    "```python\n",
    ">>> dog = Animal('Fido', 'dog')\n",
    ">>> dog.species\n",
    "'dog'\n",
    "```\n",
    "\n",
    "### Part B [10 points]:\n",
    "\n",
    "What happens when the client tries to set the value of the `species` attribute as in the following:\n",
    "```python\n",
    ">>> dog.species = 'cat'\n",
    "```\n",
    "You don't need to write the answer to this question, but you should test it.\n",
    "\n",
    "Thus to prevent the client's previous code from breaking, we need to create another method that:\n",
    "1. Allows the client to set the species of an `Animal` object using `.species`\n",
    "2. And checks that the species attribute is set to a species in our `valid_species` list.\n",
    "\n",
    "Add another method `species(self, into)` to the `Animal` class. This method should set the `_species` attribute of this `Animal` to `into`, but before that, it makes sure that `into` is in the `valid_species` list. If `into` is not in the `valid_species` list, raise an `Exception`. Apply the `species.setter` decorator to this method.\n",
    "\n",
    "If this is done correctly, then a client should still be able to set the species of an `Animal` object (within the restricted list) as follows:\n",
    "```python\n",
    ">>> dog = Animal('Fido', 'dog')\n",
    ">>> dog.species\n",
    "'dog'\n",
    ">>> dog.species = 'cat'\n",
    ">>> dog.species\n",
    "'cat'\n",
    ">>> dog.species = 'TheThing'\n",
    "...\n",
    "AssertionError: invalid species: TheThing\n",
    "```\n",
    "\n",
    "Do our clients need to change the implementation of their `transform` function now? \n",
    "Nope, and hopefully they give you a bonus!\n",
    "\n",
    "#### Deliverables\n",
    "- `animal.py`"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<a id='p2'></a>\n",
    "## Problem 2: Linked List\n",
    "\n",
    "A singly linked list can be implemented by nodes. Each node has 2 properties: a head and a tail. The head is the element held by the first node in the list. The tail is the rest of the list, which is a reference to the next node.\n",
    "\n",
    "The code below is a partially implemented singly linked list. It consists of 2 classes: `LinkedList` is a node that holds an element and `Nil` is a special node that indicates the end of the linked list. In this problem, you will finish the implementation of these 2 classes. To name a node `LinkedList` is a bit confusing, but as we will see, although the `LinkedList` is implemented as a node, it behaves as if it is a whole list.\n",
    "\n",
    "We start by copying the following template code into the file `linked_list.py`.\n",
    "\n",
    "**Hint:** Even though this problem seems long, the first 3 parts are solvable with a one-line statement and Part D is only slightly longer (depending your implementation choices). Also, you don't need any loops in this problem (you will not lose points if you do so, but you should find it quite unnatural to use loops here).\n",
    "\n",
    "```python\n",
    "class LinkedList:\n",
    "    \n",
    "    def __init__(self, head, tail):\n",
    "        assert isinstance(tail, LinkedList) or isinstance(tail, Nil), TypeError(\n",
    "            'tail should either be a LinkedList or a Nil')\n",
    "        self._head, self._tail = head, tail\n",
    "    \n",
    "    @property\n",
    "    def head(self):\n",
    "        return self._head\n",
    "    \n",
    "    @property\n",
    "    def tail(self):\n",
    "        return self._tail\n",
    "    \n",
    "    def __str__(self):\n",
    "        return str(self._head) + ' -> ' + str(self._tail)\n",
    "    \n",
    "    def __repr__(self):\n",
    "        return f'LinkedList({repr(self._head)}, {repr(self._tail)})'\n",
    "    \n",
    "    def __len__(self):\n",
    "        return 1 + len(self._tail)\n",
    "\n",
    "    def __getitem__(self, i):\n",
    "        return self._head if i == 0 else self._tail[i-1]\n",
    "    \n",
    "    def prepend(self, val):\n",
    "        pass # TODO\n",
    "    \n",
    "    def append(self, val):\n",
    "        pass # TODO\n",
    "    \n",
    "    def for_each(self, fun):\n",
    "        pass # TODO\n",
    "    \n",
    "    def summation(self):\n",
    "        return self._head + self._tail.summation() if self._tail else self._head\n",
    "    \n",
    "    def minimum(self):\n",
    "        def smaller(a, b):\n",
    "            return a if a < b else b\n",
    "        return smaller(self._head, self._tail.minimum()) if self._tail else self._head\n",
    "    \n",
    "    def reduce_right(self, fun):\n",
    "        pass # TODO\n",
    "\n",
    "class Nil():\n",
    "    \n",
    "    def __str__(self): \n",
    "        return 'Nil'\n",
    "    \n",
    "    def __repr__(self):\n",
    "        return 'Nil()'\n",
    "    \n",
    "    def __len__(self):\n",
    "        return 0\n",
    "    \n",
    "    def __getitem__(self, i):\n",
    "        raise IndexError('index out of range')\n",
    "    \n",
    "    def __bool__(self): \n",
    "        return False\n",
    "    \n",
    "    def prepend(self, val): \n",
    "        pass # TODO\n",
    "    \n",
    "    def append(self, val):  \n",
    "        return LinkedList(val, Nil())\n",
    "    \n",
    "    def for_each(self, fun):\n",
    "        pass # TODO\n",
    "```\n",
    "\n",
    "\n",
    "### Part A [15 points]: \n",
    "\n",
    "For this problem, we want to make this linked list immutable and we are not updating its head or tail after it's been created. \n",
    "\n",
    "**Please implement the `append` method.** It returns a `LinkedList` where a node holding the object `val` is appended. \n",
    "\n",
    "**Hint:** It's recursive. We recursively invoke the `append` on the tail, the tail of the tail, and so on.\n",
    "\n",
    "There is a price to pay for the immutability. If we want to append an element to this list, we will need to update the tail of the last `LinkedList` node, which breaks the immutability. Therefore, we need to re-create the last `LinkedList` node, but then we realize that we need to update the next-to-last `LinkedList` node by re-creating it as well and so on. It is essentially re-creating the whole new list with one more node. This is not very efficient. Fortunately, instead of appending an element to the end of this list, we can also prepend an element to the start of this list, which will be much more efficient.\n",
    "\n",
    "#### Please implement the `prepend` method in both the `LinkedList` and `Nil` classes.\n",
    "\n",
    "**Note:** Both `append` and `prepend` are returning a new object without mutating the object passed to it (which is our linked list, the `self`). In this problem, rather than imagining `append` and `prepend` as operations done on the original linked list, we imagine `append` and `prepend` as pure functions that map some input (original linked list) to an output (a linked list extended by a new node).\n",
    "\n",
    "\n",
    "### Part B [5 points]: \n",
    "\n",
    "We want to be able to conveniently apply a function to all elements in this list. Please implement the `for_each` method (in both `LinkedList` and `Nil`), which takes a function `fun` as an argument, and returns a `LinkedList` (or a `Nil`) in which the element `e` in the list is mapped to `fun(e)`. For example:\n",
    "\n",
    "```python\n",
    "l = Nil().prepend(1).prepend(2).prepend(3).prepend(4)\n",
    "def square(x):\n",
    "    return x**2\n",
    "print(l)\n",
    "print(l.for_each(square))\n",
    "```\n",
    "gives\n",
    "```\n",
    "4 -> 3 -> 2 -> 1 -> Nil\n",
    "16 -> 9 -> 4 -> 1 -> Nil\n",
    "```\n",
    "\n",
    "**Note:** Again, in the above example, `for_each` is not modifying the values in the `LinkedList` `l`. It is returning a new `LinkedList`. \n",
    "\n",
    "### Part C [10 points]: \n",
    "\n",
    "If we look at the logic of `summation` and `minimum`, we will notice that they are quite similar. Both are first *reducing* the tail of the list into an intermediate result and then *combining* the head with the intermediate result. The difference between `summation` and `minimum` lies in what this *combining* function is.  \n",
    "For `summation`, two elements are combined by the function\n",
    "```python \n",
    "def combine(a, b): \n",
    "    return a+b\n",
    "```\n",
    "(We don't really have this function in our code, but its logic is used in `summation`.) \n",
    "For `minimum`, two elements are combined by the hypothetical function\n",
    "```python\n",
    "def combine(a, b):\n",
    "    return a if a < b else b\n",
    "```\n",
    "(We don't really have this function in our code, but its logic is used in `minimum`.)\n",
    "Ideally, we should be able to create any `combine` function and apply it to the linked list. Please implement the `reduce_right` method, which requires a function `fun` as an argument. `reduce_right` will first *reduce* the tail of the list into an intermediate result by applying `fun` to the tail nodes. Then, `fun` should be applied to the head and the intermediate result to combine them.  \n",
    "For example:\n",
    "```python\n",
    "l = Nil().prepend(1).prepend(2).prepend(3).prepend(4)\n",
    "def smaller(a, b): # our \"combine\" function\n",
    "    return a if a < b else b\n",
    "print(l)\n",
    "print(l.reduce_right(smaller))\n",
    "```\n",
    "gives\n",
    "```\n",
    "4 -> 3 -> 2 -> 1 -> Nil\n",
    "1\n",
    "```\n",
    "\n",
    "### Part D [10 points]: \n",
    "\n",
    "For this problem, copy the following into `longest_sentence.py`:\n",
    "\n",
    "```python\n",
    "def get_list_of_sentences(chapter1='swansway-chapter1.txt'):\n",
    "    def to_sentences(p):\n",
    "            for delimiter in '.?!': p = p.replace(delimiter, '|')\n",
    "            return [s.strip('\\\" ') for s in p.split('|')]\n",
    "    with open(chapter1, 'r', encoding='UTF-8') as f:\n",
    "        paragraphs = f.readlines()\n",
    "    \n",
    "    sentences = [s for p in paragraphs for s in to_sentences(p) if len(s) > 1]\n",
    "    list_of_sentences = Nil()\n",
    "    for s in sentences[::-1]:\n",
    "        list_of_sentences = list_of_sentences.prepend(s)\n",
    "    return list_of_sentences\n",
    "\n",
    "\n",
    "def longest_sentence():\n",
    "    list_of_sentences = get_list_of_sentences()\n",
    "      \n",
    "    pass # TODO\n",
    "```\n",
    "\n",
    "Marcel Proust is known for writing long sentences. The `swansway-chapter1.txt` contains the first chapter of Remembrance of Things Past. The function `get_list_of_sentences` will preprocess this file and return a `LinkedList` of all sentences in this chapter. Your task is to implement `longest_sentence` which returns the number of words in the longest sentence. You can assume that anything delimited by a space is a word.\n",
    "\n",
    "**Note:** `get_list_of_sentences` reads text from `swansway-chapter1.txt`. You should have the text file in your `HW5-final/` directory.\n",
    "\n",
    "**Hint:** Use the `for_each` and `reduce_right` methods that you implemented.\n",
    "\n",
    "\n",
    "#### Deliverables\n",
    "- `linked_list.py`\n",
    "- `longest_sentence.py`"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<a id='p3'></a>\n",
    "## Problem 3: Binary Search Tree\n",
    "\n",
    "A binary search tree (BST) is a binary tree in which:\n",
    "* The left subtree of a node contains only nodes with keys less than the node’s key.\n",
    "* The right subtree of a node contains only nodes with keys greater than the node’s key.\n",
    "* The left and right subtree each must also be a binary search tree.\n",
    "\n",
    "The code below is a partially implemented binary search tree. `BSTNode` is a node in the binary search tree. It contains a key-value pair and has references to its child nodes. `BSTTable` is the binary search tree itself. It holds a reference to the root node of the tree and contains methods that manipulate the nodes. A `BSTTable` can be used as a table to store key-value pairs. Our implementation will only be able to handle the case when the keys are comparable to each other. Simple client code of `BSTTable` looks like this:\n",
    "\n",
    "```python\n",
    "greektoroman = BSTTable()\n",
    "greektoroman.put('Athena',    'Minerva')\n",
    "greektoroman.put('Eros',      'Cupid')\n",
    "greektoroman.put('Aphrodite', 'Venus')\n",
    "greektoroman.get('Eros')\n",
    "```\n",
    "and will return\n",
    "```\n",
    "'Cupid'\n",
    "```\n",
    "\n",
    "In this problem, we will implement the `get` and `put` method of `BSTTable`. \n",
    "\n",
    "We start by copying the following template code into the file `binary_search_tree.py`.\n",
    "\n",
    "**Hint:** You can `print` a `BSTTable` or `BSTNode` to see its structure. It's not pretty, but should be enough for debugging purposes.\n",
    "\n",
    "**Note:** You may assume that the keys are unique.\n",
    "\n",
    "```python\n",
    "class BSTNode:\n",
    "\n",
    "    def __init__(self, key, val):\n",
    "        self.key, self.val = key, val\n",
    "        self.left = None\n",
    "        self.right = None\n",
    "        self.size = 1\n",
    "    \n",
    "    def __str__(self):\n",
    "        return f'BSTNode({self.key}, {self.val})' + \\\n",
    "               '\\n|\\n|-(L)->' + '\\n|      '.join(str(self.left ).split('\\n')) + \\\n",
    "               '\\n|\\n|-(R)->' + '\\n|      '.join(str(self.right).split('\\n'))\n",
    "        \n",
    "        \n",
    "class BSTTable:\n",
    "    def __init__(self):\n",
    "        self._root = None\n",
    "    \n",
    "    def __str__(self):\n",
    "        return str(self._root)\n",
    "    \n",
    "    def __len__(self):\n",
    "        return self._size(self._root)\n",
    "        \n",
    "    def put(self, key, val):\n",
    "        self._root = self._put(self._root, key, val)\n",
    "    \n",
    "    def get(self, key):\n",
    "        return self._get(self._root, key)\n",
    "    \n",
    "    def _put(self, node, key, val):\n",
    "        pass # TODO\n",
    "    \n",
    "    def _get(self, node, key):\n",
    "        pass # TODO\n",
    "    \n",
    "    @staticmethod\n",
    "    def _size(node):\n",
    "        return node.size if node else 0\n",
    "```\n",
    "\n",
    "### Part A [15 points]: \n",
    "\n",
    "The `put` method inserts a key-value pair into the `BSTTable`. It delegates the function call to `_put`. Please implement the 'private' method `_put`. It takes a `node`, a `key`, and a `val` as arguments. `node` is the root of the subtree into which the key-value pair is to be inserted into. After the insertion, the root of this subtree may change. `_put` should return the new root of this subtree. The difference of the input subtree and the returned subtree is that in the returned subtree, a new key-value pair has been inserted. If the key was already in the subtree, then we update the value of the corresponding node.\n",
    "\n",
    "**Note:** We assume all the keys are of the same type and are comparable to each other with `>`, `<` and `==`. So if all your keys are strings, they will be ordered lexicographically. If all your keys are numbers, they will be ordered from the smallest to the largest.\n",
    "\n",
    "**Hint:** This is recursive. Think about the base case: what should be returned if the `node` passed in is `None`? Which subtree (left/right) should we insert if the `node` is not `None`? Remember to update the `size` of the nodes. \n",
    "\n",
    "### Part B [15 points]: \n",
    "\n",
    "The `get` method returns the value with the corresponding key. It delegates the function call to `_get`. Please implement the 'private' method `_get`. It takes a `node` and a `key` as an argument. It will search for a node with the same key under the subtree whose root is `node`. If there is such a node, return the `val` of this node. `_get` should raise a `KeyError` if no such node is found.\n",
    "\n",
    "\n",
    "#### Deliverables\n",
    "- `binary_search_tree.py`"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.7"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
